import db_postgres, macros, options, postgres, sequtils, strutils, times, timeutils, uuids

import nre except toSeq
from unicode import capitalize, toLower

type
  User* = object
    id*: UUID
    displayName*, email*, hashedPwd*: string

  ApiToken* = object
    id*, userId: UUID
    name*, hashedToken: string
    expires: Option[DateTime]

  Measure* = object
    id*, userId*: UUID
    slug*, name*, description*, domainUnits*, rangeUnits*: string
    domainSource*, rangeSource*: Option[string]
    analysis*: seq[string]

  Value* = object
    id*, measureId*: UUID
    value: int
    timestamp: DateTime
    extData: string

  MutateClauses = object
    columns*: seq[string]
    placeholders*: seq[string]
    values*: seq[string]

let UPPERCASE_PATTERN = re"(.)(\p{Lu})"

proc newMutateClauses(): MutateClauses =
  return MutateClauses(
    columns: @[],
    placeholders: @[],
    values: @[])

proc dbFormat(s: string): string = return s

proc dbFormat(dt: DateTime): string = return dt.formatIso8601

proc dbFormat[T](list: seq[T]): string =
  return "{" & list.mapIt("'" & dbFormat(it) & "'").join(",") & "}"

proc dbFormat[T](item: T): string = return $item

macro populateMutateClauses(t: typed, newRecord: bool, mc: var MutateClauses): untyped =

  # Must be working with an object.
  let tTypeImpl = t.getTypeImpl
  if not (tTypeImpl.typeKind == ntyObject):
    error $t & " is not an object."

  result = newStmtList()

  # iterate over all the object's fields
  for child in tTypeImpl[2].children:

    # ignore AST nodes that are not field definitions
    if child.kind == nnkIdentDefs:

        # grab the field, it's string name, and it's type
        let field = child[0]
        let fieldType = child[1]
        let fieldName = $field

        # we do not update the ID, but we do check: if we're creating a new
        # record, we should not have an existing ID
        if fieldName == "id":
          result.add quote do:
            if `newRecord` and not `t`.id.isZero:
              raise newException(
                AssertionError,
                "Trying to create a new record, but the record already has an ID.")

        # if we're looking at an optional field, add logic to check for presence
        elif fieldType.kind == nnkBracketExpr and
           fieldType.len > 0 and
           fieldType[0] == Option.getType:

          result.add quote do:
            `mc`.columns.add(`fieldName`)
            if `t`.`field`.isSome:
              `mc`.placeholders.add("?")
              `mc`.values.add(dbFormat(`t`.`field`.get))
            else:
              `mc`.placeholders.add("NULL")

        # otherwise assume we can convert and go ahead.
        else:
          result.add quote do:
            `mc`.columns.add(`fieldName`)
            `mc`.placeholders.add("?")
            `mc`.values.add(dbFormat(`t`.`field`))

  #echo result.repr

# TODO
#macro rowToRecord(recType: untyped, row: seq[string]): untyped
#  echo recType

macro recordName(rec: typed): string =
  return $rec.getTypeInst

proc tableNameForRecord[T](rec: T): string =
  return recordName(rec).replace(UPPERCASE_PATTERN, "$1_$2").toLower() & "s"

proc createRecord[T](db: DbConn, rec: T): T =
  var mc = newMutateClauses()
  populateMutateClauses(rec, true, mc)

  # Confusingly, getRow allows inserts and updates. We use it to get back the ID
  # we want from the row.
  let newIdStr = db.getValue(sql(
    "INSERT INTO " & tableNameForRecord(rec) &
    " (" & mc.columns.join(",") & ") " &
    " VALUES (" & mc.placeholders.join(",") & ") " &
    " RETURNING id"), mc.values)

  result = rec
  result.id = parseUUID(newIdStr)
  
proc updateRecord[T](db: DbConn, rec: T): bool =
  var mc = newMutateClauses()
  populateMutateClauses(rec, false, mc)

  let setClause = zip(mc.columns, mc.placeholders).mapIt(it.a & " = " it.b).join(',')
  let numRowsUpdated = db.execAffectedRows(sql(
    "UPDATE " & tableNameForRecord(rec) &
    " SET " & setClause &
    " WHERE id = ? "), mc.values.concat(@[rec.id]))

  return numRowsUpdated > 0;

# TODO
#proc getRecord[T](db: DbConn, UUID id): T =

macro listFieldNames(t: typed): untyped =
  let tTypeImpl = t.getTypeImpl

  if not (tTypeImpl.typeKind == ntyObject):
    error $t & " is not an object."

  var fieldNames: seq[tuple[n: string, t: string]] = @[]
  for child in tTypeImpl[2].children:
    if child.kind == nnkIdentDefs:
      echo $child[1]
      fieldNames.add((n: $child[0], t: $child[1]))

  result = newLit(fieldNames)


proc createUser*(db: DbConn, user: User): User = db.createRecord(user)
proc createApiToken*(db: DbConn, token: ApiToken): ApiToken = db.createRecord(token)
proc createMeasure*(db: DbConn, measure: Measure): Measure = db.createRecord(measure)
proc createValue*(db: DbConn, value: Value): Value = db.createRecord(value)

#[
when isMainModule:

  let u = User(
    displayName: "Bob",
    email: "bob@bobsco.com",
    hashedPwd: "test")

  echo createRecord(nil, u)
]#
